from cum.scrapers.foolslide import FoOlSlideChapter, FoOlSlideSeries
import re


class DokiReaderSeries(FoOlSlideSeries):
    BASE_URL = "https://kobato.hologfx.com/reader/"
    url_re = re.compile(r'https?://kobato\.hologfx\.com/reader/series/')

    def get_chapters(self):
        return super().get_chapters(DokiReaderChapter)


class DokiReaderChapter(FoOlSlideChapter):
    BASE_URL = DokiReaderSeries.BASE_URL
    url_re = re.compile(r'https?://kobato\.hologfx\.com/reader/read/')

    def from_url(url):
        return FoOlSlideChapter.from_url(url, DokiReaderSeries)



from cum.scrapers.base import BaseChapter, BaseSeries, download_pool
from functools import partial
from urllib.parse import urljoin
import concurrent.futures
import re
import requests

name_re = re.compile(r'(?P<type>Chapter|Special) (?P<num>[0-9\.]+)(?:$|\: )'
                     r'(?P<title>.*)')
fallback_re = re.compile(r'(?P<num>.*?)(?:$|\: )(?P<title>.*)')


class DynastyScansSeries(BaseSeries):
    url_re = re.compile(r'https?://dynasty-scans\.com/series/')

    def __init__(self, url, **kwargs):
        super().__init__(url, **kwargs)
        url = url.replace('http://', 'https://')
        if url.endswith('/'):
            url = url[:-1]
        jurl = url + '.json'
        self.json = requests.get(jurl).json()
        self.chapters = self.get_chapters()

    def get_chapters(self):
        chapters = []
        for t in self.json['taggings']:
            if 'permalink' in t and 'title' in t:
                name_parts = re.search(name_re, t['title'])
                if not name_parts:
                    name_parts = re.search(fallback_re, t['title'])
                    chapter = name_parts.group('num')
                elif name_parts.group('type') == 'Special':
                    chapter = 'Special ' + name_parts.group('num')
                else:
                    chapter = name_parts.group('num')
                title = name_parts.group('title')
                url = urljoin('https://dynasty-scans.com/chapters/',
                              t['permalink'])
                c = DynastyScansChapter(name=self.name, alias=self.alias,
                                        chapter=chapter, url=url, title=title)
                chapters.append(c)
        return chapters

    @property
    def name(self):
        return self.json['name']


class DynastyScansChapter(BaseChapter):
    url_re = re.compile(r'https?://dynasty-scans\.com/chapters/')
    uses_pages = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.groups:
            self.groups = self.get_groups()

    def download(self):
        data = requests.get(self.url + '.json').json()
        pages = [urljoin('https://dynasty-scans.com',
                 u['url']) for u in data['pages']]
        files = [None] * len(pages)
        futures = []
        with self.progress_bar(pages) as bar:
            for i, page in enumerate(pages):
                r = requests.get(urljoin(self.url, page), stream=True)
                fut = download_pool.submit(self.page_download_task, i, r)
                fut.add_done_callback(partial(self.page_download_finish,
                                              bar, files))
                futures.append(fut)
            concurrent.futures.wait(futures)
            self.create_zip(files)

    def from_url(url):
        url = url.replace('http://', 'https://')
        if url.endswith('/'):
            url = url[:-1]
        j = requests.get(url + '.json').json()
        author_link = None
        for t in j['tags']:
            if t['type'] == 'Series':
                series_url = urljoin('https://dynasty-scans.com/series/',
                                     t['permalink'])
                series = DynastyScansSeries(series_url)
                for chapter in series.chapters:
                    if chapter.url == url:
                        return chapter
            elif t['type'] == 'Author':
                author_link = urljoin('https://dynasty-scans.com/authors/',
                                      t['permalink'])
        if author_link:
            series = DynastyScansSeries(author_link)
            for chapter in series.chapters:
                if chapter.url == url:
                    return chapter
        # if the chapter is a one-shot
        name = j['title']
        return DynastyScansChapter(name=name, chapter='0', url=url)

    def get_groups(self):
        data = requests.get(self.url + '.json').json()
        groups = []
        for t in data['tags']:
            if t['type'] == 'Scanlator':
                groups.append(t['name'])
        return groups



from cum.scrapers.foolslide import FoOlSlideChapter, FoOlSlideSeries
import re


class YuriismSeries(FoOlSlideSeries):
    url_re = re.compile(r'https?://www\.yuri-ism\.net/slide/series/')

    @property
    def BASE_URL(self):
        if self.use_https:
            return 'https://www.yuri-ism.net/slide/'
        else:
            return 'http://www.yuri-ism.net/slide/'

    def get_chapters(self):
        return super().get_chapters(YuriismChapter)


class YuriismChapter(FoOlSlideChapter):
    BASE_URL = YuriismSeries.BASE_URL
    url_re = re.compile(r'https?://www\.yuri-ism\.net/slide/read/')

    def from_url(url):
        return FoOlSlideChapter.from_url(url, YuriismSeries)



from bs4 import BeautifulSoup
from cum import config, exceptions, output
from cum.scrapers.base import BaseChapter, BaseSeries, download_pool
from functools import partial
from mimetypes import guess_type
from urllib.parse import urljoin, urlparse
import concurrent.futures
import re
import requests


class MangadexSeries(BaseSeries):
    """Scraper for mangadex.org.

    Some examples of chapter info used by Mangadex (matched with `name_re`):
        Vol. 2 Ch. 18 - Strange-flavored Ramen
        Ch. 7 - Read Online
        Vol. 01 Ch. 001-013 - Read Online
        Vol. 2 Ch. 8 v2 - Read Online
        Oneshot
    """
    url_re = re.compile(r'(?:https?://mangadex\.(?:org|com))?/manga/([0-9]+)')
    name_re = re.compile(r'Ch\. ?([A-Za-z0-9\.\-]*)(?: v[0-9]+)?(?: - (.*))')
    language_re = re.compile(r'/images/flags/')
    group_re = re.compile(r'/group/([0-9]+)')

    def __init__(self, url, **kwargs):
        super().__init__(url, **kwargs)
        self._get_page(self.url)
        self.chapters = self.get_chapters()
        # fetch paginated chapters
        while True:
            next_url = self._get_next_url()
            if not next_url:
                break
            self._get_page(next_url)
            self.chapters = self.get_chapters() + self.chapters

    def _get_page(self, url):
        r = requests.get(url)
        self.soup = BeautifulSoup(r.text, config.get().html_parser)

    def _get_next_url(self):
        pagination = self.soup.find('ul', class_='pagination')
        active_item = (pagination.find('li', class_='active')
                       if pagination else None)
        next_item = (active_item.find_next_sibling('li', class_='paging')
                     if active_item else None)
        next_link = next_item.find('a') if next_item else None
        next_url = next_link.get('href') if next_link else None
        return urljoin('https://mangadex.com', next_url) if next_url else None

    def get_chapters(self):
        links = self.soup.find_all('a')
        chapters = []

        manga_name = self.name
        for a in links:
            url = a.get('href')
            url_match = re.search(MangadexChapter.url_re, url) if url else None
            if not url_match:
                continue
            name = a.string
            if not name:
                print("no name for {}".format(url))
                continue
            name = name.strip()
            name_parts = re.search(self.name_re, name)
            chapter = name_parts.group(1) if name_parts else name
            title = name_parts.group(2) if name_parts else None
            title = None if title == 'Read Online' else title
            language = a.parent.parent.find('img', src=self.language_re)\
                                      .get('title')
            # TODO: Add an option to filter by language.
            if language != 'English':
                continue
            groups = [a.parent.parent.find('a', href=self.group_re).string]
            c = MangadexChapter(name=manga_name, alias=self.alias,
                                chapter=chapter,
                                url=urljoin('https://mangadex.org', url),
                                groups=groups, title=title)
            chapters = [c] + chapters
        return chapters

    @property
    def name(self):
        title_re = re.compile(r'^(.+) \(Manga\) - MangaDex')
        title = self.soup.find('title').string.strip()
        return re.search(title_re, title).group(1)


class MangadexChapter(BaseChapter):
    # match /chapter/12345 and avoid urls like /chapter/1235/comments
    url_re = re.compile(
        r'(?:https?://mangadex\.(?:org|com))?/chapter/([0-9]+)'
        r'(?:/[^a-zA-Z0-9]|/?$)'
    )
    # There is an inlined js with a bunch of useful variables
    hash_re = re.compile(r'var dataurl ?= ?\'([A-Za-z0-9]{32})\'')
    # Example: (mind that the trailing comma is invalid json)
    # var page_array = [
    # 'x1.jpg','x2.jpg','x3.jpg','x4.jpg','x5.jpg','x6.png',];
    pages_re = re.compile(r'var page_array ?= ?\[([^\]]+)\]', re.DOTALL)
    # Just extract the single page name like: x1.jpg
    single_page_re = re.compile(r'\s?\'([^\']+)\',?')
    # This can be a mirror server or data path. Example:
    # var server = 'https://s2.mangadex.org/'
    # var server = '/data/'
    server_re = re.compile(r'var server ?= ?\'([^\']+)\'')
    uses_pages = True

    @staticmethod
    def _reader_get(url, page_index):
        return requests.get(url)

    def available(self):
        self.r = self.reader_get(1)
        if not len(self.r.text):
            return False
        elif self.r.status_code == 404:
            return False
        elif re.search(re.compile(r'Chapter #[0-9]+ does not exist.'),
                       self.r.text):
            return False
        else:
            return True

    def download(self):
        if getattr(self, 'r', None):
            r = self.r
        else:
            r = self.reader_get(1)
        soup = BeautifulSoup(r.text, config.get().html_parser)
        chapter_hash = re.search(self.hash_re, r.text).group(1)
        pages_var = re.search(self.pages_re, r.text)
        pages = re.findall(self.single_page_re, pages_var.group(1))
        files = [None] * len(pages)
        mirror = re.search(self.server_re, r.text).group(1)
        server = urljoin('https://mangadex.org', mirror)
        futures = []
        last_image = None
        with self.progress_bar(pages) as bar:
            for i, page in enumerate(pages):
                if guess_type(page)[0]:
                    image = server + chapter_hash + '/' + page
                else:
                    print('Unkown image type for url {}'.format(page))
                    raise ValueError
                r = requests.get(image, stream=True)
                if r.status_code == 404:
                    r.close()
                    raise ValueError
                fut = download_pool.submit(self.page_download_task, i, r)
                fut.add_done_callback(partial(self.page_download_finish,
                                              bar, files))
                futures.append(fut)
                last_image = image
            concurrent.futures.wait(futures)
            self.create_zip(files)

    def from_url(url):
        r = MangadexChapter._reader_get(url, 1)
        soup = BeautifulSoup(r.text, config.get().html_parser)
        try:
            series_url = soup.find('a', href=MangadexSeries.url_re)['href']
        except TypeError:
            raise exceptions.ScrapingError('Chapter has no parent series link')
        series = MangadexSeries(urljoin('https://mangadex.org', series_url))
        for chapter in series.chapters:
            parsed_chapter_url = ''.join(urlparse(chapter.url)[1:])
            parsed_url = ''.join(urlparse(url)[1:])
            if parsed_chapter_url == parsed_url:
                return chapter

    def reader_get(self, page_index):
        return self._reader_get(self.url, page_index)



from cum import config, output, sanity
from math import sqrt
from natsort import humansorted
from shutil import copyfile
from sqlalchemy import (
    Boolean,
    Column,
    create_engine,
    DateTime,
    ForeignKey,
    Integer,
    String,
    Table
)
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship, sessionmaker
from sqlalchemy.orm.exc import NoResultFound
from urllib.parse import urlparse
import click
import datetime
import os
import sqlalchemy.engine.url

Base = declarative_base()

group_table = Table(
    'group_association', Base.metadata,
    Column('chapter_id', Integer, ForeignKey('chapters.id')),
    Column('group_id', Integer, ForeignKey('groups.id'))
)


class Series(Base):
    __tablename__ = 'series'

    id = Column(Integer, primary_key=True)

    name = Column(String)
    alias = Column(String, unique=True, nullable=False)
    url = Column(String, unique=True)
    following = Column(Boolean, default=True)
    directory = Column(String)
    last_updated = Column(DateTime)

    chapters = relationship("Chapter", backref="series")

    def __init__(self, series):
        self.name = series.name
        self.alias = series.alias
        self.url = series.url
        self.directory = series.directory

    @staticmethod
    def alias_lookup(alias, unfollowed=False):
        """Returns a DB object for a series by alias name. Prints an error if
        an invalid alias is specified.
        """
        filters = {'alias': alias}
        if not unfollowed:
            filters['following'] = not unfollowed
        try:
            s = session.query(Series).filter_by(**filters).one()
        except NoResultFound:
            output.error('Could not find alias "{}"'.format(alias))
            exit(1)
        else:
            return s

    def check_alias_uniqueness(self):
        """Check if the series alias is unique before initalizing the series
        object. If the alias is not unique, "-X" is appended to the end of the
        alias where the X is the next available integer. However, if the alias
        is not unique and the series with the same alias is unfollowed, the
        unfollowed series gets the new alias.
        """
        alias = self.alias
        changed = False
        count = 1
        while session.query(Series).filter_by(alias=alias).all():
            changed = True
            alias = '{}-{}'.format(self.alias, count)
            count += 1
        if changed:
            series = session.query(Series).filter_by(alias=self.alias).one()
            if not series.following:
                series.alias = alias
                return
        self.alias = alias

    @property
    def last_added(self):
        """Returns the last time a chapter has been added to the series."""
        updates = sorted([x for x in self.chapters if x.added_on is not None],
                         reverse=True, key=lambda x: x.added_on)
        if updates:
            return updates[0].added_on
        else:
            return None

    def mark_as_updated(self):
        """Updates the timestamp for the last update."""
        self.last_updated = datetime.datetime.now()
        session.commit()

    @property
    def needs_update(self):
        """Returns a boolean indicating if the series is in need of an update
        based on the average release interval.
        """
        if not self.last_added:
            return True
        next_update = self.last_added + self.release_interval
        return datetime.datetime.now() >= next_update

    @property
    def ordered_chapters(self):
        return humansorted(self.chapters, key=lambda x: x.chapter)

    @property
    def release_interval(self):
        """Calculates the average second interval for the series chapter
        releases by calculating an average between new release scrapes and
        subtracting the standard deviation from that. Returns the interval as a
        timedelta object.

        Current time is added to release dates to add an interval between the
        last released chapter and current time.

        Since there is a massive number of chapters added right after each
        other when the series is first added, the interval between two chapter
        releases must be greater than 60 seconds to be counter.
        """
        release_dates = [x.added_on for x in self.chapters if x.added_on]
        release_dates.append(datetime.datetime.now())
        release_dates = sorted(release_dates)
        intervals = []
        for i in range(len(release_dates) - 1):
            interval = (release_dates[i+1] - release_dates[i]).total_seconds()
            if interval > 60:
                intervals.append(interval)
        if len(intervals) == 0:
            average_seconds = 0
        elif len(intervals) == 1:
            average_seconds = intervals[0]
        else:
            mean = float(sum(intervals)) / len(intervals)
            devsum = sum((x - mean) ** 2 for x in intervals)
            dev = sqrt(float(devsum) / len(intervals))
            average_seconds = max((mean - dev/2), 0)
        return datetime.timedelta(seconds=average_seconds)


class Chapter(Base):
    __tablename__ = 'chapters'

    id = Column(Integer, primary_key=True)
    series_id = Column(Integer, ForeignKey('series.id'))

    # Downloaded has the value -1 for ignored chapters, 0 for new chapters and
    # 1 for downloaded chapters.
    downloaded = Column(Integer, default=0)
    chapter = Column(String)
    url = Column(String, unique=True)
    title = Column(String)
    added_on = Column(DateTime)
    api_id = Column(String)

    groups = relationship('Group', secondary=group_table, backref='chapters')

    def __init__(self, chapter, series):
        self.series = series
        self.chapter = chapter.chapter
        self.title = chapter.title
        self.url = chapter.url
        self.added_on = datetime.datetime.now()
        self.api_id = getattr(chapter, 'api_id', None)

        self.groups = []
        for group in chapter.groups:
            try:
                g = session.query(Group).filter(Group.name == group).one()
            except NoResultFound:
                g = Group(group)
                session.add(g)
                session.commit()
            self.groups.append(g)

    @staticmethod
    def find_new(alias=None):
        """Return a list of new chapters as Chapter objects and applies human
        sorting to it. Accepts an optional 'alias' argument, which will filter
        the query.
        """
        query = session.query(Chapter).join(Series).filter(Series.following)
        if alias:
            query = query.filter(Series.alias == alias)
        query = query.filter(Chapter.downloaded == 0).all()
        return humansorted([x.to_object() for x in query],
                           key=lambda x: x.chapter)

    @property
    def group_tag(self):
        """Return a joined string of chapter's groups enclosed in brackets."""
        return ''.join(['[{}]'.format(x.name) for x in self.groups])

    @property
    def status(self):
        """Return the chapter's downloaded status as a one character flag."""
        if self.downloaded == 0:
            return 'n'
        elif self.downloaded == -1:
            return 'i'
        else:
            return ' '

    def to_object(self):
        """Turns a database entry into a chapter object for the respective
        site by parsing the URL.
        """
        parse = urlparse(self.url)
        kwargs = {
            'name': self.series.name,
            'alias': self.series.alias,
            'chapter': self.chapter,
            'url': self.url,
            'groups': self.groups,
            'directory': self.series.directory,
            'api_id': self.api_id
        }
        if parse.netloc in ('mangadex.com', 'mangadex.org'):
            from cum.scrapers.mangadex import MangadexChapter
            return MangadexChapter(**kwargs)
        if parse.netloc == 'bato.to':
            from cum.scrapers.batoto import BatotoChapter
            return BatotoChapter(**kwargs)
        elif parse.netloc == 'dynasty-scans.com':
            from cum.scrapers.dynastyscans import DynastyScansChapter
            return DynastyScansChapter(**kwargs)
        elif parse.netloc == 'manga.madokami.al':
            from cum.scrapers.madokami import MadokamiChapter
            return MadokamiChapter(**kwargs)
        elif parse.netloc == 'kobato.hologfx.com':
            from cum.scrapers.dokireader import DokiReaderChapter
            return DokiReaderChapter(**kwargs)
        elif parse.netloc == 'www.yuri-ism.net':
            from cum.scrapers.yuriism import YuriismChapter
            return YuriismChapter(**kwargs)
        elif parse.netloc == 'manga.famatg.com':
            from cum.scrapers.fallenangels import FallenAngelsChapter
            return FallenAngelsChapter(**kwargs)
        elif parse.netloc == 'mangasupa.com':
            from cum.scrapers.mangasupa import MangaSupaChapter
            return MangaSupaChapter(**kwargs)


class Group(Base):
    __tablename__ = 'groups'

    id = Column(Integer, primary_key=True)
    name = Column(String, unique=True)

    def __init__(self, name):
        self.name = name

    def __str__(self):
        return self.name


def backup_database():
    """Backs up the database file to a file called cum.db.bak."""
    db_path = os.path.join(config.cum_dir, 'cum.db')
    backup_path = os.path.join(config.cum_dir, 'cum.db.bak')
    copyfile(db_path, backup_path)


def initialize():
    global db_path, engine, session
    db_path = os.path.join(config.cum_dir, 'cum.db')
    db_url = sqlalchemy.engine.url.URL('sqlite', database=db_path)
    engine = create_engine(db_url)
    if not os.path.exists(db_path):
        Base.metadata.create_all(bind=engine)
    Session = sessionmaker(bind=engine)
    session = Session()


def test_database():
    """Runs a database sanity test."""
    sanity_tester = sanity.DatabaseSanity(Base, engine)
    sanity_tester.test()
    if sanity_tester.errors:
        for error in sanity_tester.errors:
            err_target, err_msg = str(error).split(' ', 1)
            message = ' '.join([click.style(err_target, bold=True), err_msg])
            output.warning(message)
        output.error('Database has failed sanity check; '
                     'run `cum repair-db` to repair database')
        exit(1)



from abc import ABCMeta
from cum import config, exceptions
from cum.scrapers.base import BaseChapter, BaseSeries, download_pool
from functools import partial
from tempfile import NamedTemporaryFile
from urllib.parse import urljoin, urlparse
import concurrent.futures
import re
import requests


class FoOlSlideSeries(BaseSeries, metaclass=ABCMeta):
    def __init__(self, url, directory=None, stub=None, use_https=False):
        self.url = url
        self.directory = directory
        self.stub = stub
        self.use_https = use_https
        self._page = 1
        self.get_comic_details()
        self.chapters = self.get_chapters()

    def _process_comic_list(self, response):
        """Iterates the JSON data provided by the list API and returns either
        the dictionary or None.
        """
        path = urlparse(self.url).path
        for comic in response['comics']:
            comic_path = urlparse(comic['href']).path
            if comic['stub'] == self.stub or comic_path == path:
                return comic

    @property
    def api_hook_details(self):
        path = 'api/reader/comic/id/{}'.format(self.foolslide_id)
        return urljoin(self.BASE_URL, path)

    @property
    def api_hook_list(self):
        path = 'api/reader/comics/page/{}'.format(self._page)
        return urljoin(self.BASE_URL, path)

    def get_comic_details(self):
        """Parses through the various series listed on Foolslide until a match
        with the specified series URL is found.
        """
        while True:
            response = requests.get(self.api_hook_list).json()
            if response.get('error', None) == 'Comics could not be found':
                raise exceptions.ScrapingError()
            result = self._process_comic_list(response)
            if result:
                break
            self._page += 1
        self.foolslide_id = result['id']
        self.name = result['name']

    def get_chapters(self, chapter_object):
        """Queries the series details API and creates a chapter object for each
        chapter listed.
        """
        response = requests.get(self.api_hook_details).json()
        chapters = []
        for chapter in response['chapters']:
            if int(chapter['chapter']['subchapter']) > 0:
                chapter_number = '.'.join([chapter['chapter']['chapter'],
                                           chapter['chapter']['subchapter']])
            else:
                chapter_number = chapter['chapter']['chapter']
            kwargs = {
                'name': self.name,
                'alias': self.alias,
                'chapter': chapter_number,
                'api_id': chapter['chapter']['id'],
                'url': chapter['chapter']['href'],
                'title': chapter['chapter']['name'],
                'groups': [team['name'] for team in chapter['teams']]
            }
            chapter = chapter_object(**kwargs)
            chapters.append(chapter)
        return chapters

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value


class FoOlSlideChapter(BaseChapter, metaclass=ABCMeta):
    uses_pages = True
    chapter_id_re = re.compile(r'"chapter_id":"([0-9]*)"')
    url_name_re = re.compile(r'/read/(.*?)/')
    no_pages_re = re.compile(r'(^.*)page.*$')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.api_id = kwargs.get('api_id')
        if self.url.split("://")[0] == "https":
            self.use_https = True
        else:
            self.use_https = False

    @property
    def api_hook_details(self):
        path = 'api/reader/chapter/id/{}'.format(self.api_id)
        return urljoin(self.BASE_URL, path)

    def download(self):
        response = requests.get(self.api_hook_details).json()
        pages = response['pages']
        files = [None] * len(pages)
        futures = []
        with self.progress_bar(pages) as bar:
            for i, page in enumerate(pages):
                r = requests.get(page['url'], stream=True)
                fut = download_pool.submit(self.page_download_task, i, r)
                fut.add_done_callback(partial(self.page_download_finish,
                                              bar, files))
                futures.append(fut)
            concurrent.futures.wait(futures)
            self.create_zip(files)

    def from_url(url, series_object):
        url = re.search(FoOlSlideChapter.no_pages_re, url).group(1)
        url_name = re.search(FoOlSlideChapter.url_name_re, url).group(1)
        series = series_object(None, stub=url_name,
                               use_https=(url.split("://")[0] == "https"))
        for chapter in series.chapters:
            if chapter.url.split("://")[1] == url.split("://")[1]:
                return chapter



from bs4 import BeautifulSoup
from cum import config, exceptions
from cum.scrapers.base import BaseChapter, BaseSeries, download_pool
from functools import partial
from mimetypes import guess_type
from urllib.parse import urljoin, urlparse
import concurrent.futures
import re
import requests


class MangaSupaSeries(BaseSeries):
    url_re = re.compile(r'http://mangasupa\.com/manga/([^.]+)$')
    title_re = re.compile(r'Read (.*) Manga Online | Mangasupa')
    from_url_re = re.compile(r'http://mangasupa.com/chapter/([^/]+)')

    def __init__(self, url, **kwargs):
        super().__init__(url, **kwargs)
        r = requests.get(self.url)
        if r.status_code != 200:
            raise exceptions.ScrapingError()
        self.soup = BeautifulSoup(r.text, config.get().html_parser)
        self.chapters = self.get_chapters()

    @property
    def name(self):
        title = self.soup.find('title').string.strip()
        return re.search(self.title_re, title).group(1)

    def get_chapters(self):
        links = self.soup.find_all('a')
        chapters = []

        id = re.search(MangaSupaSeries.url_re, self.url).group(1)
        manga_name = self.name
        for a in links:
            url = a.get('href')
            url_match = (re.search(MangaSupaChapter.url_re, url)
                         if url else None)
            if url_match and url_match.group(1) == id:
                chapter = url_match.group(2)
                title = re.search(MangaSupaChapter.title_re, a.text)\
                          .group(1).strip()
                c = MangaSupaChapter(name=manga_name, alias=self.alias,
                                     chapter=chapter,
                                     url=url,
                                     groups=[], title=title)
                chapters = [c] + chapters

        return chapters


class MangaSupaChapter(BaseChapter):
    url_re = re.compile(
        r'http://mangasupa.com/chapter/([^/]+)/chapter_([0-9]+)'
    )
    title_re = re.compile(r'(?:vol[\.0-9 ]+)?chapter[\.0-9v ]+:?(.*)$')
    alt_re = re.compile(r'page [0-9]+ - MangaSupa.com')
    uses_pages = False

    def from_url(url):
        series_id = re.search(MangaSupaSeries.from_url_re, url).group(1)
        series = MangaSupaSeries("http://mangasupa.com/manga/" + series_id)
        for chapter in series.chapters:
            if chapter.url == url:
                return chapter
        return None

    def download(self):
        if getattr(self, 'r', None):
            r = self.r
        else:
            r = requests.get(self.url)

        soup = BeautifulSoup(r.text, config.get().html_parser)
        vung_doc = soup.find("div", {"class": "vung_doc"})
        pages = vung_doc.find_all("img")
        files = [None] * len(pages)
        futures = []
        assert len(pages) > 0

        with self.progress_bar(len(pages)) as bar:
            for i, page in enumerate(pages):
                r = requests.get(page.get('src'), stream=True)
                if r.status_code == 404:
                    r.close()
                    raise ValueError
                fut = download_pool.submit(self.page_download_task, i, r)
                fut.add_done_callback(partial(self.page_download_finish,
                                              bar, files))
                futures.append(fut)

            concurrent.futures.wait(futures)
            self.create_zip(files)



from cum.scrapers.dokireader import DokiReaderSeries, DokiReaderChapter
from cum.scrapers.dynastyscans import DynastyScansChapter, DynastyScansSeries
from cum.scrapers.madokami import MadokamiChapter, MadokamiSeries
from cum.scrapers.mangadex import MangadexSeries, MangadexChapter
from cum.scrapers.yuriism import YuriismChapter, YuriismSeries
from cum.scrapers.mangasupa import MangaSupaChapter, MangaSupaSeries

series_scrapers = [
    DokiReaderSeries,
    DynastyScansSeries,
    MadokamiSeries,
    MangadexSeries,
    YuriismSeries,
    MangaSupaSeries
]
chapter_scrapers = [
    DokiReaderChapter,
    DynastyScansChapter,
    MadokamiChapter,
    MangadexChapter,
    YuriismChapter,
    MangaSupaChapter
]



from requests.exceptions import ConnectionError
from sqlalchemy.exc import IntegrityError as DatabaseIntegrityError


class CumException(Exception):
    """Base class for all cum exceptions."""

    def __init__(self, message=''):
        self.message = message

    def __str__(self):
        return repr(self.message)


class LoginError(CumException):
    pass


class ScrapingError(CumException):
    pass


class ConfigError(CumException):
    """Exception that is thrown when cum encounters a malformed configuration
    file. Accepts a string representing the raw text of the configuration file,
    the cursor position as a (row, column) tuple and a message.
    """

    def __init__(self, config, cursor, message=''):
        self.config = config
        self.cursor = cursor
        super().__init__(message)
